spinner {
  opacity: 0;
  transition: opacity 200ms ease-in-out;
  background-color: rgba(64, 64, 64, 0.5);
}

spinner.visible {
  opacity: 1;
}

.output-overlay {
  font-size: 96px;
  background-color: @theme_selected_bg_color;
  color: @theme_selected_fg_color;
  border-radius: 8px;
  opacity: 0.9;
  padding: 8px;
}

.output-overlay .description {
  font-size: 12px;
}
